from checker import checker_remove
from local_push import edge_addition_multiple
from local_push import edge_deletion_multiple
from local_push import reverse_local_push
from local_push import update_reverse_push_multi

# Counter Factual Explanation

cdef class CFE:
    # Assumption is that edges can only be deleted.

    def __init__(self, graph, p, r, double alpha, double  epsilon):
        self.graph = graph
        self.p = p
        self.r = r
        self.p_no_u = {}
        self.r_no_u = {}
        self.alpha = alpha
        self.epsilon = epsilon

    def compute_pagerank_wo_u(self, user, items):
        """
        This function computes PPR(n_i, item | A) for all the items and all the neighbors of the user  
        """
        self.r_no_u = {}
        self.p_no_u = {}
        p_no_u = dict(self.p)
        r_no_u = dict(self.r)
        deleted_neighbors = []
        for node in self.graph.successors(user):
            if node != user:
                deleted_neighbors.append(node)
        if self.graph.has_edge(user, user):
            # compute the delta r
            for item in items:
                update_reverse_push_multi(self.graph, user, deleted_neighbors, p_no_u[item], r_no_u[item],
                                          self.alpha)

            # update graph
            weights = edge_deletion_multiple(self.graph, user, deleted_neighbors)

            # recompute scores
            for item in items:
                reverse_local_push(self.graph, item, p_no_u[item], r_no_u[item], alpha=self.alpha, e=self.epsilon,
                                   update=True)
                self.r_no_u[item] = dict(r_no_u[item])
                self.p_no_u[item] = dict(p_no_u[item])
                # self.p_no_u[item], self.r_no_u[item] = reverse_local_push(self.graph, item, {}, {}, alpha=self.alpha,
                #                                                   e=self.epsilon)
        else:
            # update graph
            weights = edge_deletion_multiple(self.graph, user, deleted_neighbors)

            # recompute scores
            for item in items:
                p_no_u[item], r_no_u[item] = reverse_local_push(self.graph, item, {}, {},
                                                                alpha=self.alpha, e=self.epsilon)
                self.r_no_u[item] = dict(r_no_u[item])
                self.p_no_u[item] = dict(p_no_u[item])
                # self.p_no_u[item], self.r_no_u[item] = reverse_local_push(self.graph, item, {}, {}, alpha=self.alpha,
                #                                                   e=self.epsilon)

        # restore graph
        edge_addition_multiple(self.graph, user, deleted_neighbors, weights)

        return p_no_u, r_no_u

    def cfe_single_item_poly(self, user, top_item, item, min_number, p):
        """
        SwapOrder in the paper
        :param user:
        :param top_item:
        :param item:
        :param min_number:
        :param p:
        :return:
        """
        # returns the possibility of replacing top item with item
        # with deleting at most min_number actions
        cdef bint replaced
        cdef double sum_diff, sum_weight, weight, diff_val

        neighbors = [node for node in self.graph.successors(user)]
        deleted_neighbors = []
        replaced = False
        diff_values = []
        for neighbor in neighbors:
            if neighbor == user:
                continue
            diff_val = p[item].get(neighbor, 0) - p[top_item].get(neighbor, 0)
            weight = self.graph[user][neighbor]['weight']
            diff_val *= weight
            diff_values.append((neighbor, diff_val, weight))
        sorted_diff_values = sorted(diff_values, key=lambda x: x[1])
        sum_diff = sum([elem[1] for elem in sorted_diff_values])
        sum_weight = sum([elem[2] for elem in sorted_diff_values])

        # the approximate ppr scores (computed by reverse push) are within epsilon
        # distance from the true value, so the difference of two approximate ppr
        # scores is within 2 * epsilon distance from the true value
        if sum_diff > 2 * self.epsilon:
            try:
                message = 'top item cannot be top item %s\t%s\t%e\t%e\t%e\t%e\n' % \
                          (top_item, item, sum_diff, self.p[top_item].get(user, 0.0), self.p[item].get(user, 0.0),
                           self.epsilon * 2)
                print(message)
            except:
                pass

        for i in range(len(sorted_diff_values) - 1):
            if i + 1 > min_number:
                break
            sum_diff -= sorted_diff_values[i][1]
            sum_weight -= sorted_diff_values[i][2]
            deleted_neighbors.append(sorted_diff_values[i][0])
            # print 'deleted neighbors', deleted_neighbors, sum_diff, sum_weight
            if sum_diff > 2 * self.epsilon * sum_weight:
                replaced = True
                break
        return replaced, deleted_neighbors, sum_diff

    def cfe_single_item_poly_why_not(self, user, top_item, why_not_item, min_number, p, items, pf, rf):
        """
        SwapOrder in the paper
        :param user:
        :param top_item:
        :param item:
        :param min_number:
        :param p:
        :return:
        """
        # returns the possibility of replacing top item with item
        # with deleting at most min_number actions
        cdef bint replaced
        cdef double sum_diff, weight, diff_val

        neighbors = [node for node in self.graph.successors(user)]
        deleted_neighbors = []
        replaced = False
        diff_values = []
        for neighbor in neighbors:
            if neighbor == user:
                continue
            diff_val = p[why_not_item].get(neighbor, 0) - p[top_item].get(neighbor, 0)
            weight = self.graph[user][neighbor]['weight']
            diff_val *= weight
            diff_values.append((neighbor, diff_val, weight))
        sorted_diff_values = sorted(diff_values, key=lambda x: x[1])
        sum_diff = sum([elem[1] for elem in sorted_diff_values])

        # the approximate ppr scores (computed by reverse push) are within epsilon
        # distance from the true value, so the difference of two approximate ppr
        # scores is within 2 * epsilon distance from the true value
        if sum_diff > 2 * self.epsilon:
            try:
                message = 'top item cannot be top item %s\t%s\t%e\t%e\t%e\t%e\n' % \
                          (top_item, why_not_item, sum_diff, self.p[top_item].get(user, 0.0),
                           self.p[why_not_item].get(user, 0.0),
                           self.epsilon * 2)
                print(message)
            except:
                pass

        for i in range(len(sorted_diff_values)):
            if i + 1 > min_number:
                break
            deleted_neighbors.append(sorted_diff_values[i][0])
            # print 'deleted neighbors', deleted_neighbors, sum_diff, sum_weight
            # new_g = self.graph.copy()
            # update_forward_push_multi(new_g, user, deleted_neighbors, pf, rf,
            #                           self.alpha,self.epsilon)
            #
            # edge_deletion_multiple(new_g, user, deleted_neighbors)
            #
            # p, r = forward_local_push(new_g, user, pf, rf, alpha=self.alpha,
            #                           e=self.epsilon,
            #                           update=True)
            # p = {key: p[key] for key in p if key in items}
            #
            explanation = [(user, node) for node in deleted_neighbors]

            if checker_remove(self.alpha, self.epsilon, explanation, self.graph.copy(), why_not_item, user,
                              items):
                # if len(p) == 0:
                #     continue
                # elif max(p,key=p.get) == why_not_item:
                replaced = True
                break
        return replaced, deleted_neighbors, sum_diff