import random

import matplotlib.pyplot as plt
import networkx as nx
import numpy as np
import pandas as pd
from matplotlib.pyplot import figure
from tqdm import tqdm

from local_push import forward_local_push

def draw_interaction_graph(g, source, top_target, replacement_item, other_users=[], targets=[], explanation=[]):
    """
    draw the graph
    :param g: networkx graph
    :param source:
    :param top_target:
    :param replacement_item:
    :param other_users:
    :param targets:
    :param explanation:
    :return:
    """
    figure(figsize=(10, 10), dpi=500)
    node_colors_map = {node: 'black' for node in g.nodes()}
    node_colors_map[source] = 'red'
    for target in targets:
        node_colors_map[target] = 'green'
    for target in other_users:
        node_colors_map[target] = 'pink'
    node_colors_map[replacement_item] = 'blue'
    node_colors_map[top_target] = 'yellow'
    node_colors = [node_colors_map[node] for node in g.nodes()]

    # red edges to neighbors of a source
    # red_edges = [(source, n) for n in g.successors(source)]
    red_edges = explanation
    black_edges = [edge for edge in g.edges() if edge not in red_edges]

    # edge labels
    # edge_labels = dict([((u, v,), d['weight'])
    #                     for u, v, d in g.edges(data=True)])

    # Need to create a layout when doing
    # separate calls to draw nodes and edges
    pos = nx.bipartite_layout(g, targets + list(g.successors(source)))
    nx.draw_networkx_nodes(g, pos, cmap=plt.get_cmap('jet'),
                           node_color=node_colors, node_size=200)
    # nx.draw_networkx_labels(g, pos)
    # nx.draw_networkx_edge_labels(g, pos, edge_labels=edge_labels)
    nx.draw_networkx_edges(g, pos, edgelist=black_edges, arrows=True)
    nx.draw_networkx_edges(g, pos, width=2, edgelist=red_edges, edge_color='r', arrows=True)

    plt.axis('off')
    plt.savefig("graph.png")
    # plt.show()

def princeGraph(full_reload=False, update=False, min_out_d=None, cutoff=None, top_actors=None, worst_actors=None,
                sample_users=None, name="graph", no_u=False, complete_ppr=True, epsilon=0.0001, alpha=0.15,
                file="no_u_g"):
    '''

    :param full_reload: reload the previous graph from file if true
    :param update: update the graph from the dataset
    :param min_out_d: minimum out degree of the users in the dataset
    :param cutoff: maximum of edge between the source node and a node for it to be part of the graph
    :param top_actors: number of user to be part of the graph from the most active users
    :param worst_actors :number of user to be part of the graph from the least active users
    :param sample_users: sample the list of users to avoid one-timers and bots
    :param name: name of the folder to save the file
    :param no_u: no user-centric graph computation
    :param epsilon: epsilon parameter of the ppr computation
    :param alpha: alpha parameter of the ppr computation
    :param complete_ppr: pre-compute the ppr for all neighbors of each users

    :return:
    '''
    if file:
        name += "/%s.nx" % (file)
    elif no_u:
        name += "/no_u_g.nx"
    else:
        name += "/g.nx"

    if full_reload:
        if no_u:
            g = nx.read_gpickle(name)
            g.graph["source"] = ""
        else:
            g = nx.read_gpickle(name)
        return g, g.graph["source"], g.graph["items"], g.graph["other_users"]
    elif update:
        def json_to_series(text):
            keys, values = zip(*[item for dct in text for item in dct.items()])
            return pd.Series([values[0] + "AA"], index=[keys[0]])

        print("reading the interactions files ")
        interactions_poetry = pd.read_json('../goodreads/out/goodreads_interactions_poetry.json', lines=True)
        interactions_comics = pd.read_json('../goodreads/out/goodreads_interactions_comics_graphic.json', lines=True)
        interactions_history = pd.read_json('../goodreads/out/goodreads_interactions_history_biography.json',
                                            lines=True)

        interactions = pd.concat([interactions_poetry, interactions_comics, interactions_history], axis=0)
        interactions = interactions[interactions.is_read]
        interactions = interactions[interactions.rating >= 3]
        interactions.loc[:, "book_id"] = interactions.book_id.astype(str)

        # interactions.loc[:, "date_added"] = pd.to_datetime(interactions.date_added.replace('', np.nan).dropna(),
        #                                                    format="%a %b %d %X %z %Y")
        interactions.to_pickle("goodreads_graph.df")
        print("reading the book files ")

        poetry_items = pd.read_json('../goodreads/out/goodreads_books_poetry.json', lines=True)
        comics_items = pd.read_json('../goodreads/out/goodreads_books_comics_graphic.json', lines=True)
        history_items = pd.read_json('../goodreads/out/goodreads_books_history_biography.json', lines=True)

        poetry_items.loc[:, "category"] = "poetry"
        comics_items.loc[:, "category"] = "comics"
        history_items.loc[:, "category"] = "history"

        book_list = pd.concat([poetry_items, comics_items, history_items], axis=0)
        book_list.loc[:, "book_id"] = book_list.book_id.astype(str)

        book_list = pd.concat([book_list, book_list.authors.apply(json_to_series)], axis=1)
        book_list = book_list.drop_duplicates(["book_id"])
        book_list.to_pickle("goodreads_book_info.df")
        # interactions = interactions[interactions.user_id.isin(interactions.user_id.sample(n=100, random_state=1))]

        interaction_graph = nx.from_pandas_edgelist(interactions, "user_id", 'book_id', ['rating'],
                                                    create_using=nx.Graph())

        book_graph = nx.from_pandas_edgelist(book_list, "author_id", 'book_id', create_using=nx.Graph())
        category_graph = nx.from_pandas_edgelist(book_list, "category", 'book_id', create_using=nx.Graph())
        book_graph = nx.compose(book_graph, category_graph)
        g = nx.compose(book_graph, interaction_graph).to_directed()

        usrs = interactions.user_id.unique()
        bks = np.append(interactions.book_id.unique(), (book_list.book_id.unique()))
        athrs = book_list.author_id.unique()

        for node in usrs:
            try:
                g.nodes[node]["type"] = "user"
            except Exception as e:
                print(e)
        for node in bks:
            try:
                g.nodes[node]["type"] = "book"
                g.nodes[node]["title"] = book_list.loc[book_list.book_id == node, "title"].values[0]
                g.nodes[node]["description"] = book_list.loc[book_list.book_id == node, "description"].values[0]

            except Exception as e:
                print(e)
        for node in athrs:
            try:
                g.nodes[node]["type"] = "author"
            except Exception as e:
                print(e)

        nx.write_gpickle(g, name)
        print("created the graph")
    else:
        g = nx.read_gpickle(name)

        interactions = g.graph["interactions"]

        print("loaded the graph")

        # user sampling
    # g.remove_edges_from(('289faec03febf3c73bd5d37ffcae0d66', '1420'))
    # interactions = interactions.loc[~((interactions.user_id == "289faec03febf3c73bd5d37ffcae0d66") & (
    #         interactions.book_id == "1420")), :]
    print("sampling the users...")
    if sample_users:
        count_users = interactions.groupby("user_id").count()
        selection = count_users[(count_users.is_read > 10) & (count_users.is_read < 100)]
        users = interactions.loc[interactions.user_id.isin(selection.index), "user_id"].drop_duplicates()
        users = users.sample(n=sample_users, random_state=1)
    elif top_actors:
        count_users = interactions.groupby("user_id").count()
        selection = count_users.sort_values("is_read")[:top_actors]
        users = interactions.loc[interactions.user_id.isin(selection.index), "user_id"].drop_duplicates()
    elif worst_actors:
        count_users = interactions.groupby("user_id").count()
        selection = count_users.sort_values("is_read", ascending=False)[:worst_actors]
        users = interactions.loc[interactions.user_id.isin(selection.index), "user_id"].drop_duplicates()
    else:
        users = interactions.loc[interactions.user_id.isin(g.nodes), "user_id"].drop_duplicates()

    # target selection
    print("finding the user...")
    if no_u:
        source = ""
    else:
        id = 1
        source = users.iloc[id]
        if min_out_d:
            while g.out_degree(source) < min_out_d:
                id += 1
                source = users.iloc[id]

    # graph reduction by item sampling
    print("reducing the graph...")
    books, g, interactions, other_users = reduce_graph(g, source, interactions, users, cutoff)

    add_weights(g, mode="uniform")

    users = users[:10]
    g.graph["items"] = books
    g.graph["users"] = users
    g.graph["interactions"] = interactions
    g.graph["source"] = source
    g.graph["other_users"] = other_users

    print("saving the graph...")

    compute_pagerank(g, users, alpha=alpha, epsilon=epsilon, complete=complete_ppr)

    if no_u:
        nx.write_gpickle(g, name)
        return g, None, g.graph["items"], g.graph["other_users"]

    nx.write_gpickle(g, name)
    return g, source, books, other_users

def reduce_graph(g, source_node="", interactions=None, other_users=None, cutoff=False, debug=False):
    if debug:
        print('============ reducing the graph ============')

    # select a subgraph with a radius of the cutoff
    if interactions is None:
        g = g.subgraph(nx.single_source_shortest_path_length(g, source_node, cutoff=cutoff).keys()).copy()
        target_nodes = [item for item in g.graph["items"] if
                        (item in g.nodes) and (item not in g.successors(source_node))]
        other_users = [user for user in g.graph["other_users"] if user in g.nodes]


    # read from the interaction data
    # filter only relevant user
    # select a subgraph with a radius of the cutoff
    elif cutoff:
        # other_users = other_users.sample(min(100, len(other_users)))
        count_users = interactions.groupby("user_id").count()
        selection = count_users[(count_users.is_read > 10) & (count_users.is_read < 100)]
        users = interactions.loc[interactions.user_id.isin(selection.index), "user_id"].drop_duplicates()
        # users = users.sample(n=500, random_state=1)
        interactions = interactions[interactions.user_id.isin(users)].sort_values("date_added").groupby(
            "user_id").head(20)
        g = g.subgraph((list(interactions.book_id.values) + list(interactions.user_id.values) + list([source_node])))
        g = g.subgraph(nx.single_source_shortest_path_length(g, source_node, cutoff=cutoff).keys()).copy()

        target_nodes = interactions.loc[((interactions.book_id.isin(g.nodes)) &
                                         (~interactions.book_id.isin(g.successors(source_node)))), "book_id"].unique()
        other_users = interactions.loc[((interactions.user_id.isin(g.nodes)) &
                                        (interactions.user_id != source_node)), "user_id"].unique()

    # only filter on the included items and users around the user
    elif source_node != "":
        target_nodes = interactions.loc[((interactions.book_id.isin(g.nodes)) &
                                         (~interactions.book_id.isin(g.successors(source_node)))), "book_id"].unique()
        other_users = interactions.loc[((interactions.user_id.isin(g.nodes)) &
                                        (interactions.user_id != source_node)), "user_id"].unique()

    # only filter on the included items and users
    else:
        target_nodes = interactions.loc[(interactions.book_id.isin(g.nodes)), "book_id"].unique()
        other_users = interactions.loc[(interactions.user_id.isin(g.nodes)), "user_id"].unique()
    return target_nodes, g, interactions, other_users

def interaction_graph_generator(num_nodes, num_edges, num_neighbors, num_targets, num_neighbors_target,
                                self_loop=True):
    '''
    create a random interaction graph

    :param num_nodes: number of node to include
    :param num_edges:  number of edge to include
    :param num_neighbors:  number of neighbors to include
    :param num_targets: number of target items to include
    :param num_neighbors_target: number of neighbors to include
    :param self_loop: include self loop
    :return:
    '''
    g = nx.DiGraph()
    target_neighbors = num_neighbors_target

    # adding nodes
    for i in range(num_nodes):
        g.add_node(i)
        if self_loop:
            g.add_edge(i, i)
        if i > 0:
            neighbor = random.randint(0, i - 1)
            g.add_edge(i, neighbor)
            g.add_edge(neighbor, i)
    s = len(g.edges())
    while s <= num_edges:
        node1 = random.randint(0, num_nodes - 1)
        f = False
        while not f:
            node2 = random.randint(0, num_nodes - 1)
            if node2 != node1 and not g.has_edge(node1, node2):
                g.add_edge(node1, node2)
                g.add_edge(node2, node1)
                f = True
                s += 2

    # add source node
    node_s = num_nodes
    g.add_node(node_s)
    if self_loop:
        g.add_edge(node_s, node_s)
    potential_neighbors = [i for i in range(num_nodes)]
    random.shuffle(potential_neighbors)
    for neighbor in potential_neighbors[:num_neighbors]:
        g.add_edge(node_s, neighbor)
        if random.randint(0, 1) == 1:
            g.add_edge(neighbor, node_s)

    # add target nodes
    target_nodes = []
    potential_neighbors = [i for i in range(num_nodes)]
    for i in range(num_targets):
        node_t = num_nodes + i + 1
        g.add_node(node_t)
        target_nodes.append(node_t)
        if self_loop:
            g.add_edge(node_t, node_t)
        random.shuffle(potential_neighbors)
        for neighbor in potential_neighbors[:target_neighbors]:
            g.add_edge(node_t, neighbor)
            g.add_edge(neighbor, node_t)

    return g, node_s, target_nodes

def add_weights(g, mode='uniform', beta=0.5):
    '''
    compute the weights in a provided graph
    :param g: networkx graph
    :param mode: "similarity" or "ratings" or 'random' or 'uniform'
    :param beta: parameter of the graph creation in similarity mode
    :return:
    '''
    if mode == "similarity":
        S = g.graph['S']
        W = nx.to_pandas_adjacency(g)
        full_s = pd.concat(S, axis=1)
        full_s.columns = full_s.columns.get_level_values(1)
        w_beta = (full_s * beta + W * beta).fillna(0)
        for u, row in tqdm(w_beta.items(), total=len(w_beta)):
            for v, weight in tqdm(row.items(), total=len(row)):
                if weight != 0:
                    if g.has_edge(u, v):
                        g[u][v]['weight'] = weight
                    else:
                        g.add_edge(u, v, weight=weight)
        g.graph['beta'] = beta
    if mode == "ratings":
        for node in g.nodes():
            sum_ratings = 0
            for neighbor in g.successors(node):
                sum_ratings += g[node][neighbor]['rating']
            for neighbor in g.successors(node):
                g[node][neighbor]['weight'] = g[node][neighbor]['rating'] / sum_ratings
    if mode == 'random':
        for node in g.nodes():
            num_succ = len(list(g.successors(node)))
            random_weights = [random.randint(1, 10) for _ in range(num_succ)]
            total_weights = sum(random_weights)
            index = 0
            for neighbor in g.successors(node):
                g[node][neighbor]['weight'] = random_weights[index] / float(total_weights)
                index += 1

    if mode == 'uniform':
        for node in g.nodes():
            num_succ = len(list(g.successors(node)))
            for neighbor in g.successors(node):
                g[node][neighbor]['weight'] = 1.0 / num_succ

def compute_pagerank(g, users, alpha=0.15, epsilon=0.0001, name="g.nx", complete=False):
    '''
    pre-computation of the ppr values in the graph file
    :param g: networkx graph
    :param users: list of the user to compute ppr for
    :param alpha: ppr parameter
    :param epsilon: ppr parameter
    :param name: name of the graph file
    :param complete: compute the ppr also for neighbors nodes
    :return:
    '''
    g.graph["epsilon"] = epsilon
    g.graph["alpha"] = alpha
    for source in tqdm(users):
        # print(e)
        p, r = forward_local_push(g, source, {}, {}, alpha=alpha, e=epsilon)
        g.nodes[source]['p'] = p
        g.nodes[source]['r'] = r
        nx.write_gpickle(g, name)
    if complete:
        for source in tqdm(users):
            p, r = g.nodes[source]['p'], g.nodes[source]['r']

            # print(e)
            p_no_u = dict(p)
            r_no_u = dict(r)
            tmp_g = g.copy()

            a = list(tmp_g.edges(source))
            tmp_g.remove_edges_from(a)
            add_weights(tmp_g, "uniform")

            for u, n in a:
                p_no_u[n], r_no_u[n] = forward_local_push(tmp_g, n, {}, {}, alpha=alpha, e=epsilon)

            g.nodes[source]['p_no_u'] = p_no_u
            g.nodes[source]['r_no_u'] = r_no_u
            nx.write_gpickle(g, name)

def remove_pagerank(g, users, complete=False):
    """
    remove the ppr computation from
    :param g: networkx graph
    :param users: list of the user to remove ppr for
    :param complete: remove the ppr also for neighbors nodes
    :return:
    """
    g = g.copy()
    del g.graph["epsilon"]
    del g.graph["alpha"]
    for source in tqdm(users):
        del g.nodes[source]['p']
        del g.nodes[source]['r']
    if complete:
        for source in tqdm(users):
            del g.nodes[source]['p_no_u']
            del g.nodes[source]['r_no_u']

    return g

def prune_graph(g, type):
    """
    remove unconnected components
    :param g: networkx graph
    :param type: node type to keep
    :return:
    """
    for c in (nx.connected_components(g.to_undirected())):
        types = [g.nodes[node]["type"] for node in c]
        if type not in types:
            # print("removing => %s" % c)
            # pprint(types)
            g.remove_nodes_from(c)

def add_types(g, types=["users", "items"]):
    """
    add node type in the graph file
    :param g: networkx graph
    :param type: node type to keep
    :return:
    """
    for type in types:
        try:
            g.graph[type] = [node for node in g.graph[type] if (node in g.nodes())]
        except:
            print(type)
    g.graph["other_users"] = [node for node in g.graph["users"] if (node in g.nodes()) and
                              ((len(list(g.successors(node))) > 10) and (len(list(g.successors(node))) < 100))]

def get_category(g, item):
    """
    return the category of an item if included in the graph
    :param g: networkx graph
    :param item: requested item
    :return:
    """
    for node in g.successors(item):
        if g.nodes[node]["type"] == "category":
            return node
    return None

def get_item_consumers(g, item):
    """
    return all users that consumed an item
    :param g: networkx graph
    :param item: requested item
    :return:
    """
    u = []
    for node in g.successors(item):
        if g.nodes[node]["type"] == "user":
            u.append(node)
    return u