<div align="center">
<h1 align="center">
<img src="https://www.etis-lab.fr/wp-content/uploads/2022/01/etis_lo_SLOGAN_H_RVB.svg" width="100"  alt="CY Cergy Logo" />
<br>
Why-Not Explainable Graph Recommender
</h1>
<h3>◦ Developed with the software and tools listed below.</h3>

<p align="center">
<img src="https://img.shields.io/badge/Jupyter-F37626.svg?style&logo=Jupyter&logoColor=white" alt="Jupyter" />
<img src="https://img.shields.io/badge/Python-3776AB.svg?style&logo=Python&logoColor=white" alt="Python" />
<img src="https://img.shields.io/badge/pandas-150458.svg?style&logo=pandas&logoColor=white" alt="pandas" />
<img src="https://img.shields.io/badge/NumPy-013243.svg?style&logo=NumPy&logoColor=white" alt="NumPy" />
<img src="https://img.shields.io/badge/Markdown-000000.svg?style&logo=Markdown&logoColor=white" alt="Markdown" />
</p>
</div>

#### Author: [Hervé-Madelein Attolou](https://www.etis-lab.fr/2022/01/17/attolou-herve-madelin/) (herve-madelein.attolou@cyu.fr)</h3>

---

## 📒 Table of Contents

- [📒 Table of Contents](#-table-of-contents)
- [📍 Overview](#-overview)
- [📂 Project Structure](#project-structure)
- [🚀 Getting Started](#-getting-started)
- [👏 Acknowledgments](#-acknowledgments)

---

## 📍 Overview

Explainable recommendation systems (RS) enhance the user experience by recommending personalized content, as well as
explanations for the given recommendations to add transparency and build
up trust in the platform. Extending the notion of explainable RS, in
this paper we define Why-Not explanations for recommendations
that were expected but not returned, and propose and implement
a technique for computing Why-Not explanations in a post-hoc
manner for a graph-based RS. Our approach builds on the notion
of counterfactual explanations in the means of a set of user-rooted
edges to add or remove in the graph in order to add the missing
recommendation to the recommendation list, and provides in this
way actionable insights on the source data and their interrelations.
Our experimental evaluation on a real-world data set demonstrates
the feasibility of our proposal and reveals interesting directions for
future work.

This repository provides the tool the produce the experiment of our paper

---

## 📂 Project Structure

```bash
repo
├── README.md
├── cfe_generator.pyx
├── checker.pyx
├── eval.ipynb
├── grapher.pyx
├── local_push.pyx
├── requirements.txt
├── sample.pyx
├── setup.py
└── swapOrder.pyx

 10 files
```

---


---

## 🚀 Getting Started

### 📦 Installation

1. Clone the repository:

```sh
git clone https://git.cyu.fr/hattolou/why-not-explainable-graph-recommender.git
```

2. Change to the project directory:

```sh
cd why-not-explainable-graph-recommender
```

3. Install the dependencies:

```sh
pip install -r requirements.txt
```

4. Build the pyx files

```sh
python setup.py build_ext --inplace   
```

### 🎮 Evaluate the model with a toy dataset

```sh
jupyter nbconvert --to notebook --execute eval.ipynb
```

### 🎰 Build your graph from your dataset

1. Create a graph using networkX

```python
import networkx as nx

g = nx.DiGraph()
```

2. Add the required nodes and edges

```python
# fantasy 18
g.add_edges_from([(18, 6), (18, 7), (18, 8), (18, 9)])
# classic 19
g.add_edges_from([(19, 10), (19, 11), (19, 12), (19, 13)])
# computer science 20
g.add_edges_from([(20, 14), (20, 15), (20, 16), (20, 17)])
....

```

3. Add information about the graph as variable in the **graph**

```python
items = list(range(6, 18))
users = list(range(0, 6))
categories = list(range(18, 21))
g.graph["items"] = items
g.graph["categories"] = categories
g.graph["users"] = users
g.graph["other_users"] = users
g.graph['interactions'] = None
g.graph["source"] = ""
....

```

3. Add information about the graph as variable in the **nodes**

```python
for node in users:
    try:
        g.nodes[node]["type"] = "user"
    except Exception as e:
        print(e)
for node in items:
    try:
        g.nodes[node]["type"] = "item"

    except Exception as e:
        print(e)
for node in categories:
    try:
        g.nodes[node]["type"] = "category"

    except Exception as e:
        print(e)

```

4. Init the weights

```python
from grapher import add_weights

add_weights(g)
```

---

## 👏 Acknowledgments

> - https://github.com/azinmatin/prince

---

## SEE ALSO

> - [a demo of the paper at the ICWE 2024 conference](https://git.cyu.fr/hattolou/emigre_icwe2024) 