import itertools
import math
from pprint import pprint

import pandas as pd

from checker import checker_remove, checker_add
from grapher import add_weights
from local_push import forward_local_push, update_forward_push_multi, reverse_local_push, edge_deletion_multiple

def swapOrderWhy(g, source_node, top_node, replacement, epsilon=0.1, alpha=0.15, pf=None, rf=None, debug=False):
    if debug:
        print('============ Swap Order ============ %s <-> %s ' % (top_node, replacement))

    neighbors = list(g.successors(source_node))
    if source_node in neighbors:
        neighbors.remove(source_node)
    a = list(g.edges(source_node))

    explanation = []
    H = {}
    sum_diff = 0
    tmp_g = g.copy()

    if pf and rf:
        update_forward_push_multi(tmp_g, source_node, neighbors, pf.copy(), rf.copy(), alpha, epsilon,
                                  action="d")
    else:
        edge_deletion_multiple(tmp_g, source_node, neighbors)
        forward_local_push(tmp_g, source_node, {}, {}, alpha=alpha, e=epsilon)

    for n in neighbors:
        p, r = forward_local_push(tmp_g, n, {}, {}, alpha=alpha, e=epsilon)
        edge_weight = g[source_node][n]['weight']
        diff = edge_weight * (p.get(top_node, 0) - p.get(replacement, 0))
        H[n] = diff
        sum_diff += diff

    H = (sorted(H.items(), key=lambda x: x[1]))
    if debug:
        print("H: %s" % H)
        print("sum to go over: %s" % sum_diff)

    if sum_diff < 0:
        if debug:
            print("epsilon too big : epsilon = %s" % epsilon)
            print("source_node = %s ,top_node=%s ,replacement=%s" % (source_node, top_node, replacement))
        return g, explanation

    while (sum_diff > 0) and (len(H) > 0):
        n, diff = H.pop()
        explanation += [(source_node, n)]
        sum_diff -= diff
        if debug:
            print("explanation: %s" % explanation)
            print("explanation size: %s" % len(explanation))
            print("sum to go over: %s" % sum_diff)
            print("H: %s" % H)

    if sum_diff > 2 * epsilon:
        explanation = a

    tmp_g = g.copy()
    tmp_g.remove_edges_from(explanation)
    add_weights(tmp_g, "uniform")

    return tmp_g, explanation

def swapOrder_whynot_remove_powerset(g, source_node, top_node, replacement, target_nodes, epsilon=0.1, alpha=0.15,
                                     pf=None, rf=None, debug=False):
    if debug:
        print('============ Swap Order COMBINAISON============ %s <-> %s ' % (top_node, replacement))

    neighbors = list(g.successors(source_node))
    if source_node in neighbors:
        neighbors.remove(source_node)

    explanation = []
    H = {}
    sum_diff = 0
    tmp_g = g.copy()
    if pf and rf:
        update_forward_push_multi(tmp_g, source_node, neighbors, pf.copy(), rf.copy(), alpha, epsilon, action="d")
    else:
        edge_deletion_multiple(tmp_g, source_node, neighbors)
        p, r = forward_local_push(tmp_g.copy(), source_node, {}, {}, alpha=alpha, e=epsilon)

    # try:
    #     p_no_u, r_no_u = g.nodes[source_node]['p_no_u'], g.nodes[source_node]['r_no_u']
    #     for u, n in a:
    #         p, r = p_no_u[n], r_no_u[n]
    #         edge_weight = g[u][n]['weight']
    #         diff = edge_weight * (p.get(top_node, 0) - p.get(replacement, 0))
    #         H[n] = diff
    #         sum_diff += diff
    # except:
    for n in neighbors:
        p, r = forward_local_push(tmp_g, n, {}, {}, alpha=alpha, e=epsilon)
        edge_weight = g[source_node][n]['weight']
        diff = edge_weight * (p.get(top_node, 0) - p.get(replacement, 0))
        H[n] = diff
        sum_diff += diff

    H = (sorted(H.items(), key=lambda x: x[1]))
    if debug:
        print("H: %s" % H)
        print("sum to go over: %s" % sum_diff)

    if sum_diff < 0:
        if debug:
            print("epsilon too big : epsilon = %s" % epsilon)
            print("source_node = %s ,top_node=%s ,replacement=%s" % (source_node, top_node, replacement))
        return g, explanation
    if sum_diff < 2 * epsilon:
        if debug:
            print("epsilon too big : epsilon = %s" % (epsilon))
            print("sum_diff  = %s < 2 * epsilon = %s" % (sum_diff, 2 * epsilon))
            print("source_node = %s ,top_node=%s ,replacement=%s" % (source_node, top_node, replacement))
        return g, []

    H = dict(H)

    check_result = False
    r = 0
    while (not check_result) and (r <= len(H.keys())):
        r += 1
        combi_set = list(itertools.combinations(H.keys(), r))
        for combi in combi_set:
            explanation = [(source_node, n) for n in combi]
            new_sum = sum([H[val[1]] for val in explanation])
            if debug:
                print("sum to go over: %s" % new_sum)
                print("current explanation: %s" % explanation)
            if new_sum >= sum_diff:
                check_result = \
                    checker_remove(alpha, epsilon, explanation, g.copy(), replacement, source_node, target_nodes,
                                   debug=False, )[0]
            if check_result:
                break
            else:
                check_result = False

    # if not check_result:
    #     return g, []

    return tmp_g, explanation

def swapOrder_whynot_remove_incremental(g, source_node, top_node, replacement, target_nodes, double epsilon=0.1,
                                        double alpha=0.15,
                                        pf=None,
                                        rf=None,
                                        bint debug=False):
    cdef double sum_diff, edge_weight, diff
    cdef bint check_result
    if debug:
        print('============ Swap Order ============ %s <-> %s ' % (top_node, replacement))

    neighbors = list(g.successors(source_node))
    if source_node in neighbors:
        neighbors.remove(source_node)

    explanation = []
    H = {}
    sum_diff = 0
    tmp_g = g.copy()

    if pf and rf:
        update_forward_push_multi(tmp_g, source_node, neighbors, pf.copy(), rf.copy(), alpha, epsilon,
                                  action="d")
    else:
        edge_deletion_multiple(tmp_g, source_node, neighbors)
        forward_local_push(tmp_g, source_node, {}, {}, alpha=alpha, e=epsilon)

    # try:
    #     p_no_u, r_no_u = g.nodes[source_node]['p_no_u'], g.nodes[source_node]['r_no_u']
    #     for u, n in a:
    #         p, r = p_no_u[n], r_no_u[n]
    #         edge_weight = g[u][n]['weight']
    #         diff = edge_weight * (p.get(top_node, 0) - p.get(replacement, 0))
    #         H[n] = diff
    #         sum_diff += diff
    # except:
    for n in neighbors:
        p, r = forward_local_push(tmp_g, n, {}, {}, alpha=alpha, e=epsilon)
        edge_weight = g[source_node][n]['weight']
        diff = edge_weight * (p.get(top_node, 0) - p.get(replacement, 0))
        H[n] = diff
        sum_diff += diff

    H = (sorted(H.items(), key=lambda x: x[1]))
    if debug:
        print("H: %s" % H)
        print("sum to go over: %s" % sum_diff)

    if sum_diff < 0:
        if debug:
            print("epsilon too big : epsilon = %s" % epsilon)
            print("source_node = %s ,top_node=%s ,replacement=%s" % (source_node, top_node, replacement))
        return g, explanation

    tmp_g = None
    check_result = False

    while (len(H) > 0) and not check_result:
        n, diff = H.pop()
        explanation += [(source_node, n)]
        sum_diff -= diff
        if sum_diff < 2 * epsilon:
            tmp_g = g.copy()
            check_result = checker_remove(alpha, epsilon, explanation, tmp_g, replacement, source_node, target_nodes,
                                          debug=debug, pf=pf.copy(), rf=rf.copy())[0]
        if debug:
            print("H: %s" % H)
            print("sum to go over: %s" % sum_diff)
            print("current explanation: %s" % explanation)

    # if not check_result:
    #     return g, []

    return tmp_g, explanation

def swapOrder_why_not_add_incremental(g, source_node, top_node, replacement, target_nodes, double epsilon=0.1,
                                      double alpha=0.15,
                                      bint debug=False,
                                      pf=None, rf=None):
    cdef double sum_diff, diff
    cdef bint check_result
    if debug:
        print('============ Swap Order ADD ============ %s <-> %s ' % (top_node, replacement))

    neighbors = list(g.successors(source_node))
    if source_node in neighbors:
        neighbors.remove(source_node)

    explanation = []
    H = {}
    sum_diff = 0

    #TODO check with g and not tmp_g
    #The initial case is the case with the neigbors included
    # tmp_g = g.copy()
    # if pf and rf:
    #     update_forward_push_multi(g, source_node, neighbors, pf.copy(), rf.copy(), alpha, epsilon,
    #                               action="d")
    # else:
    #     edge_deletion_multiple(g, source_node, neighbors)
    #     forward_local_push(g, source_node, {}, {}, alpha=alpha, e=epsilon)

    for n in neighbors:
        p, r = forward_local_push(g, n, {}, {}, alpha=alpha, e=epsilon)

        diff = (p.get(top_node, 0) - p.get(replacement, 0))
        sum_diff += diff

    # get all the ppr from the replacement node
    p_r_top, r_top = reverse_local_push(g, top_node, {}, {}, alpha=alpha, e=epsilon)

    p_r_replacement, r_r = reverse_local_push(g, replacement, {}, {}, alpha=alpha, e=epsilon)

    if debug:
        pprint([("ppr(%s,%s) -> %s" % (replacement, v, p_r_replacement[v])) for v in
                sorted(p_r_replacement, key=p_r_replacement.get)])

    if debug:
        pprint([("ppr(%s,%s) -> %s" % (top_node, v, p_r_top[v])) for v in
                sorted(p_r_top, key=p_r_top.get)])

    new_sum_diff = 0
    for node in p_r_replacement:

        if not ((node in g.successors(source_node)) or
                node == source_node or
                node == replacement or
                node == top_node or
                g.nodes[node]["type"] != "item"):
            diff = (p_r_top.get(node, 0) - p_r_replacement.get(node, 0))
            H[node] = diff
            new_sum_diff += diff

    H = (sorted(H.items(), key=lambda x: x[1], reverse=True))
    if debug:
        print("H: %s" % H)
        print("sum to go over: %s" % sum_diff)
    #
    if sum_diff < 0 or (sum_diff < new_sum_diff):
        if debug:
            print("epsilon too big : epsilon = %s" % epsilon)
            print("source_node = %s ,top_node=%s ,replacement=%s" % (source_node, top_node, replacement))
        return g, explanation

    tmp_g = g.copy()
    check_result = False
    while (len(H) > 0) and not check_result:
        n, diff = H.pop()
        explanation += [(source_node, n)]
        sum_diff += diff
        if sum_diff < 0:
            tmp_g = g.copy()
            check_result = checker_add(alpha, epsilon, explanation, tmp_g, replacement, source_node, target_nodes,
                                       debug=False)[0]
        if debug:
            print("H: %s" % H)
            print("sum to go over: %s" % sum_diff)
            print("current explanation: %s" % explanation)

    # if not check_result:
    #     return g, []

    return tmp_g, explanation

def swapOrder_why_not_add_powerset(g, source_node, top_node, replacement, target_nodes, epsilon=0.1, alpha=0.15,
                                   pf=None, rf=None, debug=False):
    if debug:
        print('============ Swap Order COMBINAISON============ %s <-> %s ' % (top_node, replacement))

    neighbors = list(g.successors(source_node))
    if source_node in neighbors:
        neighbors.remove(source_node)

    explanation = []
    H = {}
    sum_diff = 0
    tmp_g = g.copy()

    if pf and rf:
        update_forward_push_multi(tmp_g, source_node, neighbors, pf.copy(), rf.copy(), alpha, epsilon,
                                  action="d")
    else:
        edge_deletion_multiple(tmp_g, source_node, neighbors)
        forward_local_push(tmp_g, source_node, {}, {}, alpha=alpha, e=epsilon)

    # try:
    #     p_no_u, r_no_u = g.nodes[source_node]['p_no_u'], g.nodes[source_node]['r_no_u']
    #     for u, n in a:
    #         p, r = p_no_u[n], r_no_u[n]
    #         edge_weight = g[u][n]['weight']
    #         diff = edge_weight * (p.get(top_node, 0) - p.get(replacement, 0))
    #         H[n] = diff
    #         sum_diff += diff
    # except:
    for n in neighbors:
        p, r = forward_local_push(tmp_g, n, {}, {}, alpha=alpha, e=epsilon)
        diff = (p.get(top_node, 0) - p.get(replacement, 0))
        sum_diff += diff

    # get all the ppr from the replacement node
    p_r_top, r_top = reverse_local_push(g, top_node, {}, {}, alpha=alpha, e=epsilon)

    p_r, r_r = reverse_local_push(g, replacement, {}, {}, alpha=alpha, e=epsilon)
    new_sum_diff = 0
    for node in p_r:

        if not ((node in g.successors(
                source_node)) or node == source_node or node == replacement or node == top_node or
                g.nodes[node]["type"] != "item"):
            diff = (p_r.get(node, 0) - p_r_top.get(node, 0))
            H[node] = diff
            new_sum_diff += diff

    H = (sorted(H.items(), key=lambda x: x[1], reverse=True))
    if debug:
        print("H: %s" % H)
        print("sum to go over: %s" % sum_diff)
    #
    if sum_diff < 0 or (sum_diff > new_sum_diff):
        if debug:
            print("epsilon too big : epsilon = %s" % epsilon)
            print("source_node = %s ,top_node=%s ,replacement=%s" % (source_node, top_node, replacement))
        return g, explanation

    # if sum_diff > 2 * epsilon:
    #     return g, []
    tmp_g = g.copy()
    check_result = False
    H = dict(H)
    r = 0
    while (not check_result) and (r <= len(H.keys())):
        r += 1
        combi_set = list(itertools.combinations(H.keys(), r))
        for combi in combi_set:
            explanation = [(source_node, n) for n in combi]
            new_sum = sum([H[val[1]] for val in explanation])
            if debug:
                print("sum to go over: %s" % new_sum)
                print("current explanation: %s" % explanation)
            if new_sum >= sum_diff:
                check_result = \
                    checker_add(alpha, epsilon, explanation, tmp_g.copy(), replacement, source_node, target_nodes,
                                debug=False, )[0]
                if check_result:
                    break
            else:
                check_result = False

    # if not check_result:
    #     return g, []

    return tmp_g, explanation

def swapOrder_whynot_remove_exhaustive(g, source_node, top_node, replacement, target_nodes, double epsilon=0.1,
                                       double alpha=0.15,
                                       pf=None,
                                       rf=None,
                                       bint debug=False, type='none', direct=False):
    cdef double sum_diff, edge_weight, diff
    cdef bint check_result
    cdef int solution_index

    if debug:
        print('============ Swap Order REMOVE NEW ============ %s <-> %s ' % (top_node, replacement))

    neighbors = list(g.successors(source_node))
    if source_node in neighbors:
        neighbors.remove(source_node)

    explanation = []
    H = {}
    sum_diff = 0
    tmp_g = g.copy()

    if pf and rf:
        update_forward_push_multi(tmp_g, source_node, neighbors, pf.copy(), rf.copy(), alpha, epsilon,
                                  action="d")
    else:
        edge_deletion_multiple(tmp_g, source_node, neighbors)
        forward_local_push(tmp_g, source_node, {}, {}, alpha=alpha, e=epsilon)

    # try:
    #     p_no_u, r_no_u = g.nodes[source_node]['p_no_u'], g.nodes[source_node]['r_no_u']
    #     for u, n in a:
    #         p, r = p_no_u[n], r_no_u[n]
    #         edge_weight = g[u][n]['weight']
    #         diff = edge_weight * (p.get(top_node, 0) - p.get(replacement, 0))
    #         H[n] = diff
    #         sum_diff += diff
    # except:

    for n in neighbors:
        p, r = forward_local_push(tmp_g, n, {}, {}, alpha=alpha, e=epsilon)
        edge_weight = g[source_node][n]['weight']
        H[n] = [edge_weight * (p.get(t, 0) - p.get(replacement, 0)) for t in target_nodes if t != replacement]
        # print("tab%s = %s" % (n, (H[n])))

        # sum_diff += diff

    HH = pd.DataFrame.from_dict(H, orient='index', columns=[t for t in target_nodes if t != replacement])
    if (HH < 0).all().any():
        if debug:
            print("H: %s" % HH)
            print('============ No Solution ============ %s <-> %s ' % (top_node, replacement))

        return g, []
    all_combinations = []

    for r in range(1, len(H.keys()), 1):
        all_combinations += (list(itertools.combinations(H.keys(), r)))

    HHH = pd.DataFrame([HH.loc[nn, :].sum(axis=0) for nn in all_combinations], index=all_combinations)

    # threshold = [pf.get(t, 0) - pf.get(replacement, 0) for t in target_nodes if t != replacement]
    # threshold = pd.Series(threshold, index=[t for t in target_nodes if t != replacement])

    if type == 'up':
        threshold = HH[HH > 0].sum()
    elif type == 'max':
        threshold = HH.max()
    else:
        threshold = HH.sum()

    HHHH = HHH.subtract(threshold)
    HHHH.loc[HHHH.sum(1).sort_values(ascending=False).index]
    margin = 2 * epsilon
    solutions = [h for h in HHHH.index if (((HHHH.T[h] > margin).all() + (HHHH.T[h] > - margin).all()))]

    if debug:
        print("H: %s" % HHHH.loc[solutions])
        print("sum to go over: %s" % threshold)

    tmp_g = g.copy()

    if direct == True:
        s = solutions[0]
        explanation = [(source_node, node) for node in s]
        if debug:
            print("H: %s" % HHHH.loc[solutions])
            print("sum to go over: %s" % sum_diff)
            print("current explanation: %s" % explanation)
    else:
        check_result = False
        solution_index = 0
        while (len(solutions) > solution_index) and not check_result:
            s = solutions[solution_index]
            explanation = [(source_node, node) for node in s]
            tmp_g = g.copy()
            if (len(explanation) > 0):
                check_result, better_options, index_replacement = checker_remove(alpha, epsilon, explanation, tmp_g,
                                                                                 replacement, source_node, target_nodes,
                                                                                 debug=debug, pf=pf.copy(),
                                                                                 rf=rf.copy())
            if debug:
                print("H: %s" % HHHH.loc[solutions])
                print("sum to go over: %s" % sum_diff)
                print("current explanation: %s" % explanation)
            solution_index += 1
        if solution_index > 1 and debug:
            print('============ WRONGGGGG ============ %s <-> %s ' % (top_node, replacement))
    # if not check_result:
    #     return g, []

    return tmp_g, explanation

def swapOrder_whynot_add_exhaustive(g, source_node, top_node, replacement, target_nodes, double epsilon=0.1,
                                    double alpha=0.15,
                                    pf=None,
                                    rf=None,
                                    bint debug=False, type='none', direct=False):
    if debug:
        print('============ Swap Order ADD NEW ============ %s <-> %s ' % (top_node, replacement))

    neighbors = list(g.successors(source_node))
    if source_node in neighbors:
        neighbors.remove(source_node)

    explanation = []
    H = {}
    sum_diff = 0

    # TODO check with g and not tmp_g
    # The initial case is the case with the neigbors included

    tmp_g = g.copy()
    if pf and rf:
        update_forward_push_multi(tmp_g, source_node, neighbors, pf.copy(), rf.copy(), alpha, epsilon,
                                  action="d")
    else:
        edge_deletion_multiple(tmp_g, source_node, neighbors)
        forward_local_push(tmp_g, source_node, {}, {}, alpha=alpha, e=epsilon)

    # get all the ppr from the replacement node
    p_r_top, r_top = reverse_local_push(g, top_node, {}, {}, alpha=alpha, e=epsilon)

    p_r_replacement, r_r = reverse_local_push(g, replacement, {}, {}, alpha=alpha, e=epsilon)

    if debug:
        pprint([("ppr(%s,%s) -> %s" % (replacement, v, p_r_replacement[v])) for v in
                sorted(p_r_replacement, key=p_r_replacement.get)])

    if debug:
        pprint([("ppr(%s,%s) -> %s" % (top_node, v, p_r_top[v])) for v in
                sorted(p_r_top, key=p_r_top.get)])

    for node in p_r_replacement:

        if not ((node in g.successors(source_node)) or
                node == source_node or
                node == replacement or
                node == top_node or
                g.nodes[node]["type"] != "item"):
            p, r = forward_local_push(g, node, {}, {}, alpha=alpha, e=epsilon)
            H[node] = [p.get(replacement, 0) - p.get(t, 0) for t in target_nodes if t != replacement]

        # print("tab%s = %s" % (n, (H[n])))

        # sum_diff += diff

    HH = pd.DataFrame.from_dict(H, orient='index', columns=[t for t in target_nodes if t != replacement])
    # if (HH < 0).all().any():
    #     if debug:
    #         print("H: %s" % HH)
    #         print('============ No Solution ============ %s <-> %s ' % (top_node, replacement))
    #
    #     return g, []

    all_combinations = []

    for r in range(1, 3, 1):
        all_combinations += (list(itertools.combinations(H.keys(), r)))

    HHH = pd.DataFrame([HH.loc[nn, :].sum(axis=0) for nn in all_combinations], index=all_combinations)
    for el in HHH.index:
        for i in el:
            try:
                HHH[i][el] = math.inf
            except KeyError as e:
                continue

    # To compute the treshold
    T = {}
    for n in neighbors:
        p, r = forward_local_push(tmp_g, n, {}, {}, alpha=alpha, e=epsilon)
        edge_weight = g[source_node][n]['weight']
        T[n] = [edge_weight * (p.get(t, 0) - p.get(replacement, 0)) for t in target_nodes if t != replacement]

    TT = pd.DataFrame.from_dict(T, orient='index', columns=[t for t in target_nodes if t != replacement])

    all_combinations = []

    for r in range(1, len(H.keys()), 1):
        all_combinations += (list(itertools.combinations(T.keys(), r)))

    TTT = pd.DataFrame([TT.loc[nn, :].sum(axis=0) for nn in all_combinations], index=all_combinations)

    if type == 'up':
        threshold = TT[TT > 0].sum()
    elif type == 'max':
        threshold = TT.max()
    else:
        threshold = TT.sum()

    HHHH = HHH.subtract(threshold)
    HHHH.loc[HHHH.sum(1).sort_values(ascending=False).index]
    margin = 2 * epsilon
    solutions = [h for h in HHHH.index if ((HHHH - margin) > 0).drop(list(h), axis=1, errors='ignore').T[h].all()]

    if debug:
        print("H: %s" % HHH.loc[solutions])
        # print("sum to go over: %s" % threshold)

    tmp_g = g.copy()

    if direct == True:
        if len(solutions) > 0:
            s = solutions[0]
            explanation = [(source_node, node) for node in s]
            tmp_g = g.copy()
            if debug:
                print("H: %s" % HHH.loc[solutions])
                print("sum to go over: %s" % sum_diff)
                print("current explanation: %s" % explanation)
    else:
        check_result = False
        solution_index = 0
        while (len(solutions) > solution_index) and not check_result:
            s = solutions[solution_index]
            explanation = [(source_node, node) for node in s]
            tmp_g = g.copy()
            if (len(explanation) > 0):
                check_result, better_options, index_replacement = checker_add(alpha, epsilon, explanation, tmp_g,
                                                                              replacement, source_node, target_nodes,
                                                                              debug=debug, pf=pf.copy(), rf=rf.copy())
            if debug:
                print("H: %s" % HHH.loc[solutions])
                print("sum to go over: %s" % sum_diff)
                print("current explanation: %s" % explanation)
            solution_index += 1
        if debug and (solution_index > 1):
            print('============ WRONG ============ %s <-> %s ' % (top_node, replacement))
    # if not check_result:
    #     return g, []

    return tmp_g, explanation

def swapOrder_whynot_brute_force(g, source_node, top_node, replacement, target_nodes, epsilon=0.1, alpha=0.15,
                                 debug=False):
    if debug:
        print('============ Swap Order BRUTE FORCE ============ %s <-> %s ' % (top_node, replacement))

    neighbors = list(g.successors(source_node))
    if source_node in neighbors:
        neighbors.remove(source_node)

    check_result = False
    r = 0
    while (not check_result) and (r <= len(neighbors)):
        r += 1
        combi_set = list(itertools.combinations(neighbors, r))
        for combi in combi_set:
            explanation = [(source_node, n) for n in combi]
            if debug:
                print("current explanation: %s" % explanation)
            check_result = checker_remove(alpha, epsilon, explanation, g.copy(), replacement, source_node, target_nodes,
                                          debug=debug)[0]
            if check_result:
                break

    # if not check_result:
    #     return g, []

    return g.copy(), explanation